package com.americancoders.edi.fixformat;

import java.util.*;
import java.text.MessageFormat;

import com.americancoders.edi.*;
/*
* Copyright (C) mendelson-e-commerce GmbH Berlin Germany 
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the Mozilla Public License
* as published by the Mozilla Software Foundation 
* http://www.mozilla.org/MPL
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* Mozilla Public License for more details.
*/

/** 
 * Tokenizer to work with flat EDI files like SAP
 * @author  S.Heller
 * @version $Revision:$
 */

public class FixformatTokenizer implements ITokenizer{

  /**ResourceBundle to localize the error messages*/
  protected ResourceBundle rb = null;
  
  /**Tokenizer to separate segments*/
  protected StringTokenizer tokenizerFileSegment = null;
  protected FixformatDataTokenizer tokenizerSegment = null;
  
  protected String actualSegmentString = null;
  
  protected TemplateSegment actualSegment = null;
  /**HashMap containing segment templates: 
	 key: segment name,
	 object: segment*/
  protected HashMap hashMapSegment = null;
    
  /**identifier length used to identify the segment name of a segment*/
  protected int segmentIdLength = 0;
  
  /**
  * @param segmentDelimiter String with characters that may separate segments
  * @param inpoutString classic EDI data to be parsed
  * @param transactionSet TransactionSet created of parsed rules file
  * @param segmentIdLength Length of the segmentId, is used to identify the actual segment
  * and to get its dataelement length.
  */
  public FixformatTokenizer( String inputString, String segmentDelimiter,
                             TransactionSet transactionSet, int segmentIdLength ){
                              
    //Load defasult LOCALE resourcebundle with fallback to english
    try{
         this.rb = ResourceBundle.getBundle( 
 			"com.americancoders.edi.fixformat.ResourceBundleFixFormatTokenizer", 
 			Locale.getDefault() );
      }
    catch ( MissingResourceException e ) {
      System.out.println( "Error: Resource bundle " 
                          + e.getClassName() + " not found." );
    }                          
                                                       
    this.segmentIdLength = segmentIdLength;
    this.tokenizerFileSegment = new StringTokenizer( inputString, segmentDelimiter );
    //get all possible segment structures
    this.hashMapSegment = this.buildHashMapSegment( transactionSet );
    this.actualSegmentString = this.tokenizerFileSegment.nextToken();
    
    this.setSegment( actualSegmentString.substring( 0, segmentIdLength ).trim());
    this.tokenizerSegment 
        = new FixformatDataTokenizer( this.actualSegment,
                                      this.actualSegmentString,
                                      this.segmentIdLength, false );
  }
  
  /**Gets the segment template of the global hashmap and sets it as actual segment
  * @param segmentId Trimed segment identifier, e.g. "E2EDK03"
  */
  protected void setSegment( String segmentId ){
    if( this.hashMapSegment.containsKey( segmentId ))
      this.actualSegment 
        = (TemplateSegment)this.hashMapSegment.get( segmentId );
    else{
      String message 
        = MessageFormat.format( this.rb.getString( "error.parsingsegment" ), 
                                new Object[]{ segmentId });
      throw new OBOEException( message );
    }
  }
  
  
 /**Builds up a hashMap containing all templates of used segments
  *as defined in the rules .xml file
  * @param transactionSet Set of templates used to get length of dataelements
  * @author S.Heller
  */
  protected HashMap buildHashMapSegment( TransactionSet transactionSet ){
    HashMap hashmap = new HashMap();
    
    Table table = transactionSet.getHeaderTable();
    
    TemplateSegment ts;
   
    if( table != null ){
        for( int i = 0; i < table.getTemplateSegmentSize(); i++ ) {
        ts = table.getTemplateSegment(i);
        hashmap.put( ts.getID(), ts);
        }               
    }
    
    table = transactionSet.getDetailTable();
    if( table != null ){
        for( int i = 0; i < table.getTemplateSegmentSize(); i++ ) {
        ts = table.getTemplateSegment(i);
        hashmap.put( ts.getID(), ts);
        }               
    }
    
    table = transactionSet.getSummaryTable();
    if( table != null ){
        for( int i = 0; i < table.getTemplateSegmentSize(); i++ ) {
        ts = table.getTemplateSegment(i);
        hashmap.put( ts.getID(), ts);
        }               
    }
    return( hashmap );
  }
    
  /**Counts the number of data elements of the actual segment
    */
    public int countDataElements(){
      return( this.actualSegment.getTemplateDESize());
    }

    /**Counts the number of segments of the file
    */
    public int countSegments(){
      return( this.tokenizerFileSegment.countTokens() );
    }

    public String getCurrentDataElement(){
      return( this.tokenizerSegment.getCurrentDataElement() );
    }

    public int getDataElementPos(){
      return(this.tokenizerSegment.getDataElementPos() );
    }

    public String getNextDataElement(){
      return( this.tokenizerSegment.nextToken() );
    }

    public IDataTokenizer getNextSegment(){
        if( !this.tokenizerFileSegment.hasMoreTokens() )
          return( null );
	this.actualSegmentString = this.tokenizerFileSegment.nextToken();
        
        this.setSegment( actualSegmentString.substring( 0, segmentIdLength ).trim());
        
        this.tokenizerSegment 
          = new FixformatDataTokenizer( this.actualSegment,
                                        this.actualSegmentString,
                                        this.segmentIdLength, true );
        return ( this.tokenizerSegment );    
    }

    public int getSegmentPos(){
		return( 1 );
    }

    public boolean hasMoreDataElements(){
	return( this.tokenizerSegment.hasMoreTokens() 
          || this.tokenizerFileSegment.hasMoreTokens() );
    }

    public boolean hasMoreSegments(){
	return( this.tokenizerFileSegment.hasMoreTokens() );
    }

    public boolean isThereADataElement(){
	return( true );
    }
    
    public IDataTokenizer resetSegment(){
	return( new FixformatDataTokenizer( this.actualSegment,
                                            this.actualSegmentString,
                                            this.segmentIdLength, true ));
    }

  
    public IDataTokenizer makeSubfieldTokenizer() {return null;}
      
    
    /**Tokenizer to separate the data elements of a single 
	 * segment by looking up the DE length in the templated
	 * of every DE, as defined in the .xml rules file
	 * @author S.Heller
	 */
    public class FixformatDataTokenizer implements IDataTokenizer{
      protected TemplateSegment segment = null;
      protected int dePositionInSegment = -1;
      protected int charPositionInSegment = 0;
      protected String currentDataElement = null;
      protected String actualSegmentString = null;
      
      protected boolean nextCall = false;
      
	  /**@param segment Segment Template
	   *@param actualSegmentString The current segment as String, used
	   * to get the DE values
	   */
      public FixformatDataTokenizer( TemplateSegment segment, 
                                     String actualSegmentString,
                                     int initialCharPosition,
                                      boolean nextCall ){
        this.nextCall = nextCall;
        this.segment = segment;
        this.charPositionInSegment = initialCharPosition;
        this.actualSegmentString = actualSegmentString;
        this.currentDataElement 
          = actualSegmentString.substring( 0, segmentIdLength ).trim();
      }
      
      public String nextToken(){
        if( nextCall ){
          this.nextCall = false;
          return( this.currentDataElement );
        }
        

        this.dePositionInSegment++;
        int maxLength = this.getCurrentDataElementMaxLength();
        //no more Token available
        if( maxLength == -1 )
          return( null );
        this.currentDataElement 
          = this.actualSegmentString.substring( this.charPositionInSegment,
                                                this.charPositionInSegment
                                                + maxLength );
        this.charPositionInSegment += maxLength;
        return( this.currentDataElement );
      }
      
      public boolean hasMoreElements(){
        return( this.dePositionInSegment 
              < this.segment.getTemplateDESize());
      }
      
      public boolean hasMoreTokens(){
        return( this.hasMoreElements());
      }

      public int countTokens(){
        return( this.segment.getTemplateDESize());
      }
      
      /** This is a fixformat, so minlength == maxlength (!)
      */
      protected int getCurrentDataElementMaxLength(){
       if( this.dePositionInSegment >= this.segment.getTemplateDESize())
        return( -1 );
        
       TemplateDE templateDE 
           = this.segment.getTemplateDE( this.dePositionInSegment );
       return( templateDE.getMaxLength() );
      }
    
    
      public String getCurrentDataElement(){
       return( this.currentDataElement );
      }

      public int getDataElementPos(){
       return( this.charPositionInSegment );
      }

    }
  
}
